(*
This file is part of the Kind verifier

* Copyright (c) 2007-2009 by the Board of Trustees of the University of Iowa, 
* here after designated as the Copyright Holder.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the University of Iowa, nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)


(** Kind lustre verifier.  Main entry point. *)

open Types
open Exceptions
open Channels
open Globals


let solver = Globals.my_solver

(********************************************************************)
let print_copyright () =
  print_string (
   "\nKind Lustre code verifier (version "^Globals.version
  ^", build "^Globals.build^")\n"
  ^"* Copyright (c) 2007-2009 by the Board of Trustees of the University of Iowa.\n"
  ^"* All rights reserved.\n"
  ^"* \n"
  ^"* Redistribution and use in source and binary forms, with or without\n"
  ^"* modification, are permitted provided that the following conditions are met:\n"
  ^"*     * Redistributions of source code must retain the above copyright\n"
  ^"*       notice, this list of conditions and the following disclaimer.\n"
  ^"*     * Redistributions in binary form must reproduce the above copyright\n"
  ^"*       notice, this list of conditions and the following disclaimer in the\n"
  ^"*       documentation and/or other materials provided with the distribution.\n"
  ^"*     * Neither the name of the University of Iowa, nor the\n"
  ^"*       names of its contributors may be used to endorse or promote products\n"
  ^"*       derived from this software without specific prior written permission.\n"
  ^"*\n"
(*  
  ^"* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY\n"
  ^"* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED\n"
  ^"* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE\n"
  ^"* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY\n"
  ^"* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES\n"
  ^"* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;\n"
  ^"* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND\n"
  ^"* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT\n"
  ^"* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS\n"
  ^"* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n"
*)
  )


(********************************************************************)
(* SIGALARM = internal or external timeout *)
(* SIGTERM = external memout *)
let catch_alarm x =
  print_to_user_final (solver#get#cc^"Signal caught: Timeout. Terminating.\n");
  exit x

let catch_term x =
  print_to_user_final (solver#get#cc^"Signal caught: Memory out. Terminating.\n");
  exit x

let catch_pipe x = 
  print_to_user_final (solver#get#cc^"Signal caught: Broken pipe. The solver has unexpectedly closed.  Terminating.\n"
  ^(match !Flags.set_my_solver with
        YICES -> ""
      | YICES_WRAPPER -> 
          solver#get#cc^"It is possible a formula exceeded the interactive yices buffer.\n"
      | CVC3 -> (if !Flags.abs_mode then 
                   solver#get#cc^"It is possible CVC3 failed to instantiate a counterexample.\n"
                   ^solver#get#cc^"Try re-running with abstraction turned off.\n" else "")
  )
  );
  exit x


(********************************************************************)
let start fn =
  begin
    Flags.do_scratch := false;
  try
    (* set up some termination conditions *)
    (* 
       Sys.set_signal Sys.sigterm (Sys.Signal_handle catch_term);
    Sys.set_signal Sys.sigalrm (Sys.Signal_handle catch_alarm);
    Sys.set_signal Sys.sigpipe (Sys.Signal_handle catch_pipe);
       *)
    begin
      match !Flags.set_my_solver with
        YICES ->
          solver#set (new Solver_yices_cmdline.solver_yices_cmdline)
      | YICES_WRAPPER ->
	  begin
	    if(Solvers_path.yicesw_path = "no") then
	      (
		print_to_user_final("\nWarning: Kind will use the command line version of Yices.\n\n");
		solver#set (new Solver_yices_cmdline.solver_yices_cmdline)
	      )else(
		solver#set (new Solver_yices.solver_yices)
	      );
	      end
      | CVC3 -> 
          solver#set (new Solver_cvc3.solver_cvc3)
      end;
      Flags.commentchar := solver#get#cc;
      if !Flags.do_scratch then
        begin
          main_ch := open_out (fn^"."^solver#get#file_extension);
          main2_ch := 
            if !Flags.separate_solvers then
              open_out (fn^"."^solver#get#file_extension^"_2")
            else
              !main_ch;
          check_ch := open_out (fn^"."^solver#get#file_extension^".ck");
        end;
      Kind_seq_loop.main fn 
  with Reachable_space_explored ->
    print_to_user_final (solver#get#cc^"+++ The property holds. +++\n");
    print_to_user_final (solver#get#cc^"Done.\n")
  | No_next_refinement_step ->
      print_to_user_final (solver#get#cc^"error: no next refinement step\n")
  end;
;;
 (* exit 0*)
  
    


(********************************************************************)
let _ =
  let usage_msg = "Performs k-induction on lustre program.\nUsage: kind [options] input_file" in
  let speclist = [
                   ("-node", Arg.String (fun x -> Flags.user_specified_main_node_name := x),
		     "\tCONTROL: User specified main node (defalut: the last node that contains a PROPRRTY)");
                   ("-n",
                    Arg.Int (fun x -> Flags.maxloops := x),
                    "\tCONTROL: Stop after n iterations (default 200)"
                   );
                   ("-bmc",
                    Arg.Unit (fun () -> Flags.do_k_induction := false),
                    "\tCONTROL: Only performs BMC (default: false)"
                   );
                   ("-step_size",
                    Arg.Int (fun x -> Flags.k_step_size := x),
                    "\tCONTROL: Only check induction every n steps (default 1)"
                   );
                   ("-sal_ind",
                    Arg.Int (fun x -> Flags.k_step_size := x+1;
                                      Flags.maxloops := x+1),
                    "\tCONTROL: Perform sal-like induction behavior at step n (default false)"
                   );
                   ("-cvc3",
                    Arg.Unit (fun () -> Flags.set_my_solver := CVC3),
                    "\tSOLVER: Use CVC3 as solver (default: yicesw)"
                   );
                   ("-yices_exec",
                    Arg.Unit (fun () -> Flags.set_my_solver := YICES),
                    "\tSOLVER: Use Yices executable as solver (default: yices)"
                   );
		   ("-z3",
                    Arg.Unit (fun () -> Flags.set_my_solver := Z3;
				        Flags.enabled_z3 :=true;
					Flags.solverflags:= "-si -smt2 MODEL=true"),
                    "\tSOLVER: Use Z3 as solver (default: yicesw)"
                   );
                   ("-flags",
                    Arg.String (fun x -> Flags.solverflags := x),
                    "\tSOLVER: Pass string to solver cammand line (default: n/a)"
                   );
                   ("-single_solver",
                    Arg.Unit (fun () -> Flags.separate_solvers := false),
                    "\tSOLVER: Uses 1 solver instance for ground & symbolic formulas (default: false)"
                   );
                   ("-separate_solvers",
                    Arg.Unit (fun () -> Flags.separate_solvers := true),
                    "\tSOLVER: Uses 2 solver instance for ground & symbolic formulas (default: true)"
                   );
                   ("-define_mod_div",
                    Arg.Unit (fun () -> Flags.define_mod_div := true),
                    "\tSOLVER: Include definitions for mod & div (may be incomplete) (default: false)"
                   );
                   ("-buffer_limit",
                    Arg.Int (fun x -> Flags.buffer_limit := x),
                    "\tSOLVER: Try decreasing this if the program hangs early (default 200000)"
                   );

                   ("-bv",
                    Arg.Unit (fun () -> Flags.bv_mode := true),
                    "\tDEF: Use bitvector representation for bool terms (cvc only, default: false)"
                   );

                   ("-aggdef",
                    Arg.Unit (fun () -> Flags.var_defs_mode := false),
                    "\tDEF: Use aggregate definitions (default: false)"
                   );
                   ("-debug",
                    Arg.Unit (fun () -> Flags.debug := true;
                                        Flags.loud := true;
                                        Flags.final_cex_loud := true),
                    "\tOUTPUT: Include debugging output (default: false)"
                   );
                   ("-quiet",
                    Arg.Unit (fun () -> Flags.loud := false;
                                        Flags.final_cex_loud := false),
                    "\tOUTPUT: Only return final answer (default: true)"
                   );
                   ("-verbose",
                    Arg.Unit (fun () -> Flags.loud := true;
                                        Flags.final_cex_loud := true),
                    "\tOUTPUT: Return more details (default: false)"
                   );
		   ("-induct_cex",
                    Arg.Unit (fun () -> Flags.inductive_cs := true),
                    "\tOUTPUT: Prints the current inductive counterexample (default: false)"
                   );
		    ("-scratch",
                    Arg.Unit (fun () -> Flags.do_scratch := true),
                    "\tOUTPUT: Do not log scratchwork to disk (default: Don't log to disk)"
                   );
                   ("-only_inputs",
                    Arg.Unit (fun () -> Flags.print_all_vars := false),
                    "\tOUTPUT: Only print inputs in counterexamples (default: false)"
                   );
                   ("-dot",
                    Arg.Unit (fun () -> Flags.print_dot_one := true;
                                        Flags.abs_mode := true;
                                        Flags.var_defs_mode := true;
                                        Flags.loud := false),
                    "\tOUTPUT: Prints dot graph of dependencies (default: false)"
                   );
                   ("-dotall",
                    Arg.Unit (fun () -> Flags.print_dot_all := true;
                                        Flags.abs_mode := true;
                                        Flags.var_defs_mode := true),
                    "\tOUTPUT: Outputs all abstraction dependency graphs to graphXXX.dot (default: false)"
                   );
                   ("-no_inlining",
                    Arg.Unit (fun () -> 
                         Flags.aggressive_inlining := 0;
                         Flags.inlining_mode := false),
                    "\tOPT: Do not attempt variable inlining (default: false)"
                   );
                   ("-aggressive_inlining",
                    Arg.Int (fun x -> 
                         Flags.aggressive_inlining := x;
                         Flags.inlining_mode := true),
                    "\tOPT: Perform aggressive inlining up to complexity n (default: 0)"
                   );
                   ("-abstract_pres",
                    Arg.Unit (fun () -> Flags.abstr_pre := true),
                    "\tOPT: Abstract terms below a pre (default: false)"
                   );
                   ("-abstract_ites",
                    Arg.Unit (fun () -> Flags.abstr_ite := true),
                    "\tOPT: Abstract ITE terms (default: false)"
                   );
                   ("-abstract_control",
                    Arg.Unit (fun () -> Flags.abstr_bool := true;
                                        Flags.abstr_ite := true),
                    "\tOPT: Abstract ITE and boolean terms (default: false)"
                   );

                   ("-loop",
                    Arg.Unit (fun () -> Flags.loopfree_mode := true;
                                        Flags.termination_check := true),
                    "\tCOMPR: Include loopfree constraint formulas + term check (default: false)"
                   );
                   ("-compression",
                    Arg.Unit (fun () -> Flags.loopfree_mode := true),
                    "\tCOMPR: Include path compression but not term check (default: false)"
                   );
                   ("-finite_term",
                    Arg.Unit (fun () -> Flags.termination_check := true),
                    "\tCOMPR: Include the finite termination check constraint, but no loopfree constraint (default: false)"
                   );
                   ("-static_loop",
                    Arg.Unit (fun () -> Flags.loopfree_mode := true;
                                        Flags.termination_check := true;
                                        Flags.static_path_compression.(base_abstr) <- true;
                                        Flags.static_path_compression.(step_abstr) <- true
                                        ),
                    "\tCOMPR: Use a statically calculated loopfree constraint (default: false)"
                   );
                   ("-initial_compression",
                    Arg.Unit (fun () -> Flags.initial_compression := true),
                    "\tCOMPR: Include compression against initial configs (default: false)"
                   );
                   ("-no_check_compression",
                    Arg.Unit (fun () -> Flags.check_compression := false),
                    "\tCOMPR: Check compression is fully defined (incomplete!) (default: false)"
                   );
                   ("-interleave_term_checks",
                    Arg.Unit (fun x -> Flags.interleave_termination_checks := true),
                    "\tCOMPR: Perform term checks every after a step (default: false)"
                   );
                   ("-compression_in_checker",
                    Arg.Unit (fun () -> Flags.compression_in_checker := true),
                    "\tCOMPR: Adds constraints to abstraction checker (default: false)"
                   );

                   ("-pre_refine_state",
                    Arg.Unit (fun () -> Flags.pre_refine_state_vars := true),
                    "\tABS: Refine state vars initially (default: false)"
                   );
                   ("-abs",
                    Arg.Unit (fun () -> Flags.abs_mode := true;
                                        Flags.var_defs_mode := true;
                                        Flags.checker_mode := true),
                    "\tABS: Use fine-grained abstraction refinement (default: false)"
                   );
                   ("-core",
                    Arg.Unit (fun () -> Flags.abs_mode := true;
                                        Flags.var_defs_mode := true;
                                        Flags.checker_mode := true;
                                        Flags.core_mode := true),
                    "\tABS: Use core-based abstraction refinement (default: false)"
                   );

                   ("-path",
                    Arg.Unit (fun () -> Flags.abs_mode := true;
                                        Flags.var_defs_mode := true;
                                        Flags.checker_mode := true;
                                        Flags.path_mode := true),
                    "\tABS: Use path-based abstraction refinement (default: false)"
                   );

                   ("-subtree",
                    Arg.Unit (fun () -> Flags.abs_mode := true;
                                        Flags.var_defs_mode := true;
                                        Flags.checker_mode := true;
                                        Flags.full_subtree_mode := true),
                    "\tABS: Use subtree-based abstraction refinement (default: false)"
                   );

                   ("-incr",
                    Arg.Unit (fun () -> Flags.abs_mode := true;
                                        Flags.var_defs_mode := true;
                                        Flags.checker_mode := true;
                                        Flags.incr_mode := true),
                    "\tABS: Use increasingly coarser-grained refinement (default: false)"
                   );

                   ("-no_inductive_check",
                    Arg.Unit (fun () -> Flags.no_inductive_check_mode := true),
                    "\tABS: Do not check refinement in inductive step (adds incompleteness, but often faster) (default: false)"
                   );

                   ("-force_refinement",
                    Arg.Int (fun x -> Flags.force_refinement := x;
                                      Flags.compression_in_checker := false),
                    "\tABS: Periodically force a refinement step with frequency n (default max_int)"
                   );
                   ("-version",
                      Arg.Unit (fun () -> print_copyright()),
                      "Print version & copyright information"
                   )
                 ]
  in
  Arg.parse speclist start usage_msg 

